/*  This source file is (c) 2010 - 2012 Volker Schatz
    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details)              */

/*! \file

    \brief Implementation of simplified low-level API

    Inter-process communication (with LED daemon, toolbar and pagebar) is
    performed by one function opening the channel on demand and other
    functions sending an appropriate message.  The framebuffer mapping and
    screen update functions are just wrappers.  Argument and return value
    constants are defined in iliad_api.h.
*/

#include <sys/mman.h>
#include <fcntl.h>
#include <liberipc/eripc.h>
#include <liberipc/eripcbusyd.h>
#include <liberipc/eripctoolbar.h>
#include <liberipc/eripcviewer.h>
#include <liberdm/display.h>
#include <liberdm/erdm.h>

#include "iliad_api.h"


/*!
    \brief IPC channel for communication with busy LED daemon

    The IPC client is started when this function is first called, the cached
    channel data is returned on subsequent calls.

    \return NULL indicates that the IPC client could not be started.
*/
erClientChannel_t busyd_channel()
{
    static erClientChannel_t channel;
    static int connected= 0;

    if( !connected ) {
        if( erIpcStartClient(ER_BUSYD_CHANNEL, &channel) )
            return NULL;
        connected= 1;
    }
    return channel;
}


/*!
    \brief Set state of busy LED

    \param state LED_OFF, LED_BLINK or LED_ON
    \return 0 on success, ILIAD_ERR_RANGE state argument out of range,
    		ILIAD_ERR_CONN could not connect to busy LED daemon,
		ILIAD_ERR_ERLIB error returned by iRex library function call
*/
int set_led(int state)
{
    erClientChannel_t channel;

    if( state < ccBusyState_Off || state > ccBusyState_Shutdown )
        return ILIAD_ERR_RANGE;
    channel= busyd_channel();
    if( !channel )
        return ILIAD_ERR_CONN;
    return busySetBusy(channel, state) ? ILIAD_ERR_ERLIB : 0;
}



/*!
    \brief IPC channel for communication with toolbar

    The IPC client is started when this function is first called, the cached
    channel data is returned on subsequent calls.

    \return NULL indicates that the IPC client could not be started.
*/
erClientChannel_t toolbar_channel()
{
    static erClientChannel_t channel;
    static int connected= 0;

    if( !connected ) {
        if( erIpcStartClient(ER_TOOLBAR_CHANNEL, &channel) )
            return NULL;
        connected= 1;
    }
    return channel;
}


/*!
    \brief Set icons to show on toolbar

    Creates toolbar icon set from scratch.

    \param iconlist Pointer to integer array indicating desired icons.  The
    least significant byte must contain the icon ID according to the
    e_iconID_t enum, the next more significant byte the icon state according to
    the e_iconState_t enum.  The list must be concluded with a -1.

    \return 0 on success, ILIAD_ERR_CONN if the IPC client could not be started,
    ILIAD_ERR_ERLIB if one of the iRex library functions returned an error.
*/
int init_icons(int *iconlist)
{
    erClientChannel_t channel= toolbar_channel();
    int *icon;
    int status= 0;

    if( !channel )
        return ILIAD_ERR_CONN;
    status |= tbSelectIconSet(channel, UAID);
    status |= tbDisableUpdate(channel, UAID);
    status |= tbClearIconSet(channel, UAID);
    for( icon= iconlist; *icon != -1; ++icon ) {
        status |= tbAppendPlatformIcon(channel, UAID, *icon & 0xFF, ccVwrToolbar);
        /* judging from ipdf sources, signal == -1 would be "no signal" */
        status |= tbSetStatePlatformIcon(channel, UAID, *icon & 0xFF, (*icon >> 8) & 0xFF);
    }
    status |= tbEnableUpdate(channel, UAID);
    return status ? ILIAD_ERR_ERLIB : 0;
}



/*!
    \brief IPC channel for communication with pagebar

    The IPC client is started when this function is first called, the cached
    channel data is returned on subsequent calls.

    \return NULL indicates that the IPC client could not be started.
*/
erClientChannel_t pagebar_channel()
{
    static erClientChannel_t channel;
    static int connected= 0;

    if( !connected ) {
        if( erIpcStartClient(ER_PAGEBAR_CHANNEL, &channel) )
            return NULL;
        connected= 1;
    }
    return channel;
}


/*!
    \brief Initialise pagebar

    This function resets the page bar, sets the number of pages to display,
    the current page, the position on the current page (if applicable) and the
    page bar mode.

    \param npages Number of pages the page bar should display
    \param page Current page, 1-based
    \param pagepos If >= 0, this is the position on the current page in
    percent, which is displayed as a narrow black bar at the bottom of the
    rounded rectangle representing the current page.
    \param zoom Determines what the pagebar displays when the number of pages
    is so large that not all of them can be displayed as rounded rectangles.
    If 1, only the pages neighbouring the current one are displayed, all as
    rounded rectangles.  If 0, the current, first and last pages are displayed
    as rectangles, while those in between are indicated by a series of vertical
    lines.

    \return 0 on success, ILIAD_ERR_CONN if IPC client could not be started,
    ILIAD_ERR_ERLIB if one of the iRex library functions returned an error.
*/
int init_pagebar(int npages, int page, int pagepos, int zoom)
{
    erClientChannel_t channel= pagebar_channel();
    int status= 0;

    if( !channel )
        return ILIAD_ERR_CONN;
    status |= pbReset(channel, UAID);
    status |= pbSetPageCount(channel, UAID, npages);
    status |= pbSetCurrentPage(channel, UAID, page);
    if( pagepos >= 0 )
        status |= pbSetCurrentPageOffset(channel, UAID, pagepos);
    status |= pbSetZoomMode(channel, UAID, zoom);
    status |= pbRedraw(channel, UAID);
    return status ? ILIAD_ERR_ERLIB : 0;
}


/*!
    \brief Set current page and position for pagebar

    \param page Current page, 1-based
    \param pagepos If >= 0, this is the position on the current page in
    percent, which is displayed as a narrow black bar at the bottom of the
    rounded rectangle representing the current page.

    \return 0 on success, ILIAD_ERR_CONN if IPC client could not be started,
    ILIAD_ERR_ERLIB if one of the iRex library functions returned an error.
*/
int set_pagebar_page(int page, int pagepos)
{
    erClientChannel_t channel= pagebar_channel();
    int status= 0;

    if( !channel )
        return ILIAD_ERR_CONN;
    status |= pbSetCurrentPage(channel, UAID, page);
    if( pagepos >= 0 )
        status |= pbSetCurrentPageOffset(channel, UAID, pagepos);
    status |= pbRedraw(channel, UAID);
    return status ? ILIAD_ERR_ERLIB : 0;
}



static unsigned char *fb0_map;

/*!
    \brief Map screen framebuffer device into memory

    Opens and maps the framebuffer device /dev/fb0 on demand and henceforth
    returns a pointer to the mapped region.  The organisation of the
    framebuffer is one byte per pixel, row major.

    \return Pointer to mapped screen space, or NULL on error
*/
unsigned char *framebuffer()
{
    int fd;

    if( !fb0_map ) {
        fd= open("/dev/fb0", O_RDWR);
        if( fd < 0 )
            return NULL;
        fb0_map= (unsigned char *)mmap(NULL, SCREEN_WIDTH*SCREEN_HEIGHT,
                              PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
        close(fd);
        if( fb0_map == MAP_FAILED )
            fb0_map= NULL;
    }
    return fb0_map;
}


/*!
    \brief Unmap screen framebuffer device from memory
*/
void unmap_framebuffer()
{
    munmap(fb0_map, SCREEN_WIDTH*SCREEN_HEIGHT);
    fb0_map= NULL;
}



/*!
    \brief Update the display with 4-bit quality within 0.1 seconds.

    This is a convenience function which merely calls update_display(DISP_4BIT,
    DISP_01SEC).
*/
void update_now4bit()
{
    update_display(DISP_4BIT, DISP_01SEC);
}


/*!
    \brief Update the iLiad's display.

    The reason for the existence of these two parameters is undocumented by
    iRex.  Possibly higher qualities and priorities are more power-intensive.

    \param quality Quality (bit accuracy) of the update
    \param priority Priority determining the delay until the update will be
    performed (this function returns immediately in any case)
    \return 0 OK, ILIAD_ERR_RANGE when one of the arguments is out of legal
    range
*/
int update_display(int quality, int priority)
{
    if( quality < 0 || quality > 2 || priority < 0 || priority > 3 )
        return ILIAD_ERR_RANGE;
    if( fb0_map )
        msync(fb0_map, SCREEN_WIDTH*SCREEN_HEIGHT, MS_SYNC);
    dmDisplay(priority, quality);
    return 0;
}


/*!
    \brief Clears the display within 0.1 seconds.
*/
void clear_display()
{
    dmDisplayEraseToWhite(dmCmdPriorHigh);
}

